from __future__ import absolute_import, annotations

import itertools as it
from typing import Any, Iterable, List, Optional, TypeVar, Union, cast

import attr

from kslurm.args.arg import Arg, DuplicatePolicy
from kslurm.args.help import print_help
from kslurm.exceptions import (
    CommandLineError,
    CommandLineErrorGroup,
    MandatoryArgError,
    TailError,
    ValidationError,
)

T = TypeVar("T", bound=Union[dict[str, Arg[Any, Any]], type])
S = TypeVar("S")


@attr.frozen
class ScriptHelp:
    script_name: str = ""
    docstring: str = ""
    usage_suffix: str = ""


def parse_args(
    args: Iterable[str],
    models: list[Arg[Any, Any]],
    escalate_errors: bool = False,
    help: ScriptHelp = ScriptHelp(),
    allow_unknown: bool = True,
    terminate_on_unknown: bool = True,
):
    try:
        return _parse_args(
            args,
            models,
            help,
            allow_unknown=allow_unknown,
            terminate_on_unknown=terminate_on_unknown,
        )
    except CommandLineError as err:
        print(err.msg)
        if isinstance(err, CommandLineErrorGroup):
            [print(e.msg) for e in err.sub_errors]
        if escalate_errors:
            raise err
        else:
            exit()


def _parse_args(
    args: Iterable[str],
    models: list[Arg[Any, Any]],
    help: ScriptHelp,
    allow_unknown: bool = True,
    terminate_on_unknown: bool = True,
):
    model_list = sorted(models, key=lambda arg: arg.priority, reverse=True)

    parsed, tail = _match_args(args, model_list, terminate_on_unknown)

    show_help = False
    for arg in parsed:
        if arg.id == "help":
            show_help = arg.value
            break

    if show_help:
        print_help(help.script_name, models, help.docstring, help.usage_suffix)
        exit(0)

    validation_errs = filter(None, [arg.validation_err for arg in parsed])

    if tail and not allow_unknown:
        tail_err = [TailError(f"{tail} does not match any args")]
    else:
        tail_err = []

    mandatory_arg_errs = cast(List[MandatoryArgError], [])
    for arg in parsed:
        try:
            arg.value
        except MandatoryArgError as err:
            mandatory_arg_errs.append(err)

    all_err = list(it.chain(validation_errs, tail_err, mandatory_arg_errs))

    if all_err:
        raise CommandLineErrorGroup("Problems found!", all_err)

    return parsed, tail


def _match_args(
    args: Iterable[str],
    arg_list: List[Arg[Any, Any]],
    terminate_on_unknown: bool = True,
):

    group_size: int = 0
    prev_arg: Optional[Arg[Any, Any]] = None
    terminated = False

    updated_args: Iterable[Optional[Arg[Any, Any]]] = []
    next_update = updated_args

    skip: list[Arg[Any, Any]] = list()
    matched: list[Arg[Any, Any]] = list()
    tail: list[str] = []

    for arg in args:
        if terminated:
            tail.append(arg)
            continue

        next_arg = None
        for argtype in arg_list:
            if argtype.match(arg) and argtype not in skip:
                if prev_arg and group_size and prev_arg.greediness > argtype.priority:
                    try:
                        next_arg = prev_arg.with_values(
                            it.chain(prev_arg.values, [arg]), True
                        )
                    except ValidationError as err:
                        next_arg = prev_arg.with_err(err)
                    group_size -= 1
                    next_update = it.chain(updated_args, [next_arg])
                    break
                updated_args = next_update
                try:
                    next_arg = argtype.with_value(arg, True)
                except ValidationError as err:
                    next_arg = argtype.with_err(err)
                group_size = argtype.num
                if argtype.terminal:
                    terminated = True
                if next_arg.duplicates == DuplicatePolicy.SKIP:
                    skip.append(argtype)
                matched.append(argtype)
                next_update = it.chain(updated_args, [next_arg])
                break
        if next_arg is None:
            if prev_arg and group_size:
                try:
                    next_arg = prev_arg.with_values(
                        it.chain(prev_arg.values, [arg]), True
                    )
                except ValidationError as err:
                    next_arg = prev_arg.with_err(err)
                group_size -= 1
                next_update = it.chain(updated_args, [next_arg])
            else:
                tail.append(arg)
                if terminate_on_unknown:
                    terminated = True

        if next_arg is not None:
            prev_arg = next_arg

    unchanged = [arg for arg in arg_list if arg not in matched]

    updated_args = it.chain(next_update, unchanged)

    return list(filter(None, updated_args)), tail



from __future__ import absolute_import

import importlib.resources as impr
from pathlib import Path

import attr

import neuroglia_helpers
from kslurm.args import Subcommand, flag, subcommand
from kslurm.args.command import command
from kslurm.cli import kbatch, kjupyter, kpy, krun
from kslurm.cli.config import config
from kslurm.installer.installer import install

NAME = "kslurm"
HOME_DIR = "KSLURM_HOME"
ENTRYPOINTS = ["kbatch", "krun", "kjupyter", "kslurm", "kpy"]


@command(typer=True)
def _neuroglia_helpers(show_src: bool = flag(["--src-dir"])):
    if show_src:
        print(Path(neuroglia_helpers.__file__).parent)
    else:
        with impr.path("kslurm.bin", "neuroglia-helpers.sh") as path:
            print(f"\nsource {path.resolve()}")


@attr.s(auto_attribs=True)
class KslurmModel:
    command: Subcommand = subcommand(
        commands={
            "kbatch": kbatch,
            "krun": krun,
            "kjupyter": kjupyter,
            "kpy": kpy,
            "config": config,
            "update": lambda x: None,  # type: ignore
            "neuroglia-helpers": _neuroglia_helpers,
        },
    )


@command
def main(args: KslurmModel, tail: list[str]) -> int:
    name, func = args.command
    entry = f"kslurm {name}"
    if name == "update":
        install(tail, NAME, HOME_DIR, ENTRYPOINTS)
    return func([entry, *tail])


if __name__ == "__main__":
    main(["kslurm", "kbatch", "-h"])



from __future__ import absolute_import

import hashlib
import importlib.resources as impr
import os
import re
import shutil
import subprocess as sp
import sys
import tarfile
import tempfile
from pathlib import Path
from typing import Any, Literal, Optional, Union, cast, overload

import attr
from virtualenv.create import pyenv_cfg  # type: ignore

from kslurm.args import Subcommand, flag, keyword, positional, shape, subcommand
from kslurm.args.command import CommandError, command
from kslurm.args.types import WrappedCommand
from kslurm.shell import Shell
from kslurm.venv import KpyIndex, MissingPipdirError, VenvCache


def _get_hash(item: Union[str, bytes]):
    if isinstance(item, str):
        item = item.encode()
    return hashlib.md5(item).hexdigest()


def _pip_freeze(venv_dir: Path):
    return sp.run(
        [venv_dir / "bin" / "python", "-m", "pip", "freeze"], capture_output=True
    ).stdout


def _get_unique_name(index: KpyIndex, stem: str = "venv", i: int = 0) -> str:
    if i == 0:
        candidate = stem
    else:
        candidate = f"{stem}{i}"
    if candidate in index:
        return _get_unique_name(index, stem, i + 1)
    return candidate


class MissingSlurmTmpdirError(CommandError):
    pass


@overload
def _get_slurm_tmpdir(allow_missing: Literal[True] = ...) -> Optional[Path]:
    ...


@overload
def _get_slurm_tmpdir(allow_missing: Literal[False] = ...) -> Path:
    ...


def _get_slurm_tmpdir(allow_missing: bool = True):
    if not os.environ.get("SLURM_TMPDIR"):
        if not allow_missing:
            raise MissingSlurmTmpdirError(
                "This command can only be used in a compute node. Use `krun` to start "
                "an interactive session"
            )
        return
    return Path(os.environ["SLURM_TMPDIR"])


@command
def _bash():
    """Echo script for inclusion in .bashrc

    e.g.
        kpy bash >> $HOME/.bashrc
    """
    with impr.path("kslurm.bin", "bash.sh") as path:
        print(f"\nsource {path.resolve()}")


@command(typer=True)
def _load(
    name: str = positional(default="", help="Test help"),
    new_name: list[str] = keyword(match=["--as"]),
    script: list[str] = keyword(match=["--script"]),
):
    """Load a saved python venv

    Run without name to list available venvs for loading
    """
    slurm_tmp = _get_slurm_tmpdir()
    if slurm_tmp:
        index = KpyIndex(slurm_tmp)

        label = new_name[0] if new_name else name
        if label in index:
            print(
                f"An environment called '{label}' already exists. You can load "
                f"'{name}' under a different name using --as:\n"
                f"\tkpy load {label} --as <name>\n"
                f"You can also activate the existing '{label}' using\n"
                f"\tkpy activate {label}"
            )
            return 1
        venv_dir = Path(tempfile.mkdtemp(prefix="kslurm-venv-", dir=slurm_tmp / "tmp"))
    else:
        index = None
        label = name
        venv_dir = Path(tempfile.mkdtemp(prefix="kslurm-"))

    venv_cache = VenvCache()

    if not name or name not in venv_cache:
        print("Valid venvs:\n" + str(venv_cache))
        return

    print(f"Unpacking venv '{name}'", end="")
    if label != name:
        print(f" as '{label}'")
    else:
        print()
    with tarfile.open(venv_cache[name], "r") as tar:
        tar.extractall(venv_dir)

    print("Updating paths")
    with (venv_dir / "bin" / "activate").open("r") as f:
        lines = f.read().splitlines()
        # \x22 and \x27 are " and ' char
        subbed = [
            re.sub(
                r"(?<=^VIRTUAL_ENV=([\x22\x27])).*(?=\1$)",
                str(venv_dir),
                line,
            )
            for line in lines
        ]
    with (venv_dir / "bin" / "activate").open("w") as f:
        f.write("\n".join(subbed))

    for root, _, files in os.walk(venv_dir / "bin"):
        for file in files:
            path = Path(root, file)
            if path.is_symlink() or not os.access(path, os.X_OK):
                continue
            with path.open("r") as f:
                lines = f.read().splitlines()
                subbed = [
                    re.sub(
                        r"(?<=^#!)/.*$",
                        str(venv_dir / "bin" / "python"),
                        line,
                    )
                    for line in lines
                ]
            with path.open("w") as f:
                f.write("\n".join(subbed))

    cfg: Any = pyenv_cfg.PyEnvCfg.from_folder(venv_dir)  # type: ignore
    cfg.update({"prompt": label, "state_hash": _get_hash(_pip_freeze(venv_dir))})
    cfg.write()

    if index is not None:
        index[label] = str(venv_dir)
        index.write()

    shell = Shell.get()
    if script:
        with Path(script[0]).open("w") as f:
            f.write(shell.source(venv_dir))
        return 2
    shell.activate(venv_dir)


@command(typer=True)
def _save(name: str = positional(), force: bool = flag(match=["--force", "-f"])):
    """Save current venv"""
    if not os.environ.get("VIRTUAL_ENV"):
        print(
            "No active virtual env detected. Please activate one, or ensure "
            "$VIRTUAL_ENV is being set correctly"
        )
    venv_cache = VenvCache()

    delete = False
    if name in venv_cache:
        if force:
            delete = True
        else:
            print(f"{name} already exists. Run with -f to force overwrite")
            return

    dest = venv_cache.get_path(name)

    _, tmp = tempfile.mkstemp(prefix="kslurm-", suffix="tar.gz")

    venv_dir = Path(os.environ["VIRTUAL_ENV"])
    cfg: Any = pyenv_cfg.PyEnvCfg.from_folder(venv_dir)  # type: ignore
    cfg.update(
        {
            "prompt": name,
            "state_hash": _get_hash(_pip_freeze(venv_dir)),
        }
    )
    cfg.write()
    with tarfile.open(tmp, mode="w:gz") as tar:
        tar.add(venv_dir, arcname="")

    if delete:
        os.remove(dest)
    slurm_tmp = _get_slurm_tmpdir()
    if slurm_tmp:
        index = KpyIndex(slurm_tmp)
        index[name] = str(venv_dir)
        index.write()
    shutil.move(tmp, dest)


@command(typer=True)
def _create(
    name: str = positional("", help="Name of the new venv"),
    version: str = shape(
        default="",
        match=lambda s: bool(re.match(r"^[23]\.\d{1,2}$", s)),
        syntax="(2|3).x",
        examples=["2.7", "3.8"],
        help="Python version to use in new venv. An appropriate executable must be on "
        "the $PATH (e.g. 3.7 -> python3.7",
    ),
    script: list[str] = keyword(match=["--script"]),
):
    """Create a new venv

    If no name provided, a placeholder name will be generated
    """
    if version:
        ver = ["-p", version]
    else:
        try:
            data = sp.run(
                "eval $($LMOD_CMD bash list python)", shell=True, capture_output=True
            )
            if match := re.search(r"(?<=python\/)\d\.\d{1,2}", data.stdout.decode()):
                ver = ["-p", match[0]]
            else:
                ver = []
        except RuntimeError:
            ver = []

    slurm_tmp = _get_slurm_tmpdir()
    if slurm_tmp:
        index = KpyIndex(slurm_tmp)
        name = name if name else _get_unique_name(index, "venv")
        if name in index:
            print(
                f"An environment called '{name}' already exists. You can activate "
                f"the existing '{name}' using\n"
                "\tkpy activate {name}"
            )
            return 1

        venv_dir = tempfile.mkdtemp(prefix="kslurm-venv-", dir=slurm_tmp / "tmp")
        no_download = ["--no-download"]
        no_index = ["--no-index"]
    else:
        index = None
        name = name if name else "venv"
        venv_dir = tempfile.mkdtemp(prefix="kslurm-")
        no_download = []
        no_index = []

    try:
        sp.run(
            [
                sys.executable,
                "-m",
                "virtualenv",
                venv_dir,
                "--symlinks",
                "--prompt",
                name,
                *ver,
                *no_download,
            ],
        )
        sp.run(
            [
                os.path.join(venv_dir, "bin", "python"),
                "-m",
                "pip",
                "install",
                "--upgrade",
                "pip",
                *no_index,
            ],
        )
    except RuntimeError as err:
        print(err.args[0])
        return 1
    if index is not None:
        index[name] = str(venv_dir)
        index.write()

    shell = Shell.get()
    if script:
        with Path(script[0]).open("w") as f:
            f.write(shell.source(Path(venv_dir)))
        return 2
    shell.activate(Path(venv_dir))


@command(typer=True)
def _activate(name: str = positional(""), script: list[str] = keyword(["--script"])):
    """Activate a venv already created or loaded

    Only works on compute nodes. Use kpy create or kpy load --as on a login node
    """
    slurm_tmp = _get_slurm_tmpdir(False)
    index = KpyIndex(slurm_tmp)
    if not name:
        print(str(index))
        return

    if name not in index:
        print(
            f"An environment with the name '{name}' has not yet been initialized. ",
            end="",
        )
        try:
            venv_cache = VenvCache()
            if name in venv_cache:
                print(
                    f"The saved environment called '{name}' can be loaded using\n"
                    f"\tkpy load {name}\n"
                )
        except MissingPipdirError:
            pass
        print(f"A new environment can be created using\n\tkpy create {name}")
        print(f"Currently initialized environments:\n{index}")
        return 1

    shell = Shell.get()
    if script:
        with Path(script[0]).open("w") as f:
            f.write(shell.source(Path(index[name])))
        return 2
    shell.activate(Path(index[name]))


@command
def _list():
    """List all saved venvs.

    To list initialized venvs (either created or loaded), run `kpy activate` without any
    arguments
    """
    venv_cache = VenvCache()
    print(str(venv_cache))
    return


@command
def _refresh():
    try:
        dir = Path(os.environ["VIRTUAL_ENV"])
        cfg: Any = pyenv_cfg.PyEnvCfg.from_folder(dir)  # type: ignore
        state_hash = cfg["state_hash"]
    except KeyError:
        return

    try:
        hsh = _get_hash(_pip_freeze(dir))
        if hsh != state_hash and cfg["prompt"][0] != "*":
            cfg["prompt"] = "*" + cfg["prompt"]
        elif hsh == state_hash and cfg["prompt"][0] == "*":
            cfg["prompt"] = cfg["prompt"][1:]
        cfg.write()
    except KeyError:
        return


def _kpy_wrapper(argv: list[str] = sys.argv):
    with impr.path("kslurm.bin", "kpy-wrapper.sh") as path:
        print(path)


@command(typer=True)
def _rm(name: str = positional("")):
    try:
        venv_cache = VenvCache()
    except MissingPipdirError as err:
        print(err.msg)
        return 1

    if not name:
        print("Valid venvs:\n" + str(venv_cache))

    if name not in venv_cache:
        print(f"{name} is not a valid venv. Currently saved venvs are:\n{venv_cache}")
        return 1

    os.remove(venv_cache[name])
    return


@attr.frozen
class _KpyModel:
    command: Subcommand = subcommand(
        commands={
            "load": _load,
            "save": _save,
            "bash": _bash,
            "create": _create,
            "activate": _activate,
            "list": _list,
            "rm": _rm,
            "_refresh": _refresh,
            "_kpy_wrapper": cast(WrappedCommand, _kpy_wrapper),
        },
    )


@command
def kpy(cmd_name: str, args: _KpyModel, tail: list[str]):
    """Set of commands for interacting with python virtual envs"""
    name, func = args.command
    entry = f"{cmd_name} {name}"
    return func([entry, *tail])


if __name__ == "__main__":
    kpy(["kpy", "--help"])



# pyright: reportPrivateUsage=false
from __future__ import absolute_import

from pathlib import Path

from pytest_mock import MockerFixture

import kslurm.installer.installer as installer


class TestMakeEnv:
    def test_creates_new_venv(self, mocker: MockerFixture, tmp_path: Path):
        envbuilder = mocker.patch.object(installer, "EnvBuilder")
        assert installer._make_env(tmp_path) == tmp_path / "venv"
        envbuilder.assert_called_once()

    def test_doesnt_create_venv_if_update_and_venv_exists(
        self, mocker: MockerFixture, tmp_path: Path
    ):
        envbuilder = mocker.patch.object(installer, "EnvBuilder")
        mocker.patch.object(installer, "_is_venv", return_value=True)
        assert installer._make_env(tmp_path, True) == tmp_path / "venv"
        envbuilder.assert_not_called()

    def test_creates_venv_if_venv_exists_but_update_not_true(
        self, mocker: MockerFixture, tmp_path: Path
    ):
        envbuilder = mocker.patch.object(installer, "EnvBuilder")
        mocker.patch.object(installer, "_is_venv", return_value=True)
        assert installer._make_env(tmp_path) == tmp_path / "venv"
        envbuilder.assert_called_once()

    def test_creates_venv_if_update_but_no_existing_venv(
        self, mocker: MockerFixture, tmp_path: Path
    ):
        envbuilder = mocker.patch.object(installer, "EnvBuilder")
        assert installer._make_env(tmp_path, True) == tmp_path / "venv"
        envbuilder.assert_called_once()


def test_install_library(mocker: MockerFixture, tmp_path: Path):
    sub = mocker.patch("subprocess.run")
    installer._install_library("pkg==3.2", tmp_path)
    sub.assert_called_once()



from __future__ import absolute_import, annotations

from typing import Callable, Dict, List, Optional, TypeVar, Union

import attr
import docstring_parser as doc
from rich.text import Text

from kslurm.args.arg import AbstractHelpTemplate, Arg, DuplicatePolicy, HelpRow
from kslurm.args.types import WrappedCommand
from kslurm.exceptions import ValidationError

T = TypeVar("T")
S = TypeVar("S")


@attr.frozen
class ShapeArg(AbstractHelpTemplate):
    title = "Shape Args"
    header = ["", "Syntax", "", "Examples", "Default", ""]
    right_align_cols = 2

    syntax: str
    examples: list[str]

    def _syntax_format(self, syntax: str):
        lines = [f"[cyan bold]{line.strip()}[/]" for line in syntax.split(" | ")]
        return "\n".join(lines)

    def usage(self, name: str, help: str, default: Optional[str]) -> str:
        return rf"[hot]\[{name}][/]"

    def row(self, name: str, help: str, default: Optional[str]):
        return [
            Text(name, style="bold"),
            self._syntax_format(self.syntax),
            ("➔" if self.examples else ""),
            Text("\n".join(self.examples)),
            Text(str(default), style="default_col") if default is not None else "",
            Text(help),
        ]


@attr.frozen
class PositionalArg(AbstractHelpTemplate):
    title = "Positional Args"
    header = []

    def usage(self, name: str, help: str, default: Optional[str]) -> str:
        if default is None:
            return f"[cyan]{name}[/]"
        return rf"[cyan]\[{name}][/]"

    def row(self, name: str, help: str, default: Optional[str]):
        return [
            Text(name, style="bold"),
            Text(default if default is not None else "REQUIRED", style="default_col"),
            Text(help),
        ]


@attr.frozen
class SubcommandTemplate(AbstractHelpTemplate):
    title = "Commands"
    header = []

    commands: dict[str, WrappedCommand]

    def usage(self, name: str, help: str, default: Optional[str]) -> str:
        return f"[cyan]{name}[/]"

    def row(
        self, name: str, help: str, default: Optional[str]
    ) -> Union[list[HelpRow], HelpRow]:
        return [
            ["", help],
            [],
            *(
                [
                    Text(name, style="bold"),
                    Text(doc.parse(func.__doc__ or "").short_description or ""),
                ]
                for name, func in self.commands.items()
            ),
        ]


# @attr.frozen
# class KeywordArg(AbstractHelpEntry):
#     title = "Keyword Args"
#     header = ["", "Default", ""]

#     name: list[str]
#     value_name: Optional[str]
#     default: str
#     help: str

#     @property
#     def usage(self):
#         return ""

#     def row(self) -> list[Union[Text, str]]:
#         return [
#             Text(", ".join(self.name)) + (
#                 Text(f" <{(self.value_name)}>", style="grey")
#                 if self.value_name else ''
#             ),
#             Text(self.default, style="default_col"),
#             Text(textwrap.fill(self.help, 70)),
#         ]


# @attr.frozen
# class ChoiceArg(AbstractHelpEntry):
#     title = "Choice Args"
#     header = ["", "Choices", "Default", ""]

#     name: str
#     choices: list[str]
#     default: str
#     help: str

#     @property
#     def usage(self):
#         return ""

#     def row(self) -> list[Union[Text, str]]:
#         return [
#             Text(self.name, style="bold"),
#             Text(",".join(self.choices), style="bold"),
#             Text(self.default, style="default_col"),
#             Text(textwrap.fill(self.help, 70)),
#         ]


def positional(
    default: Optional[str] = None,
    *,
    help: str = "",
    name: str = "",
    format: Callable[[str], T] = str,
) -> Arg[T, None]:
    return Arg[T, None](
        priority=0,
        match=lambda _: True,
        duplicates=DuplicatePolicy.SKIP,
        format=format,
        help=help,
        help_template=PositionalArg(),
        name=name,
    ).with_value(default)


def choice(
    match: List[str],
    *,
    default: Optional[str] = None,
    help: str = "",
    name: str = "",
    format: Callable[[str], T] = str,
) -> Arg[T, None]:
    def check_match(val: str) -> T:
        if val in match:
            return format(val)
        choices = "\n".join([f"\t\t• {m}" for m in match])
        raise ValidationError(f"Please select between:\n" f"{choices}")

    return Arg[T, None](
        match=lambda _: True,
        priority=0,
        duplicates=DuplicatePolicy.SKIP,
        format=check_match,
        help=help,
        name=name,
    ).with_value(default)


Subcommand = tuple[str, WrappedCommand]


def subcommand(
    commands: Dict[str, WrappedCommand],
    default: Optional[str] = None,
) -> Arg[Subcommand, None]:
    def check_match(val: str):
        if val in commands.keys():
            return (val, commands[val])
        choices = "\n".join([f"\t\t• {m}" for m in commands.keys()])
        raise ValidationError(f"Please select between:\n" f"{choices}")

    return Arg[Subcommand, None](
        match=lambda _: True,
        priority=0,
        duplicates=DuplicatePolicy.SKIP,
        format=check_match,
        help="Run any command followed by -h for more information",
        help_template=SubcommandTemplate(commands),
        name="subcommand",
        terminal=True,
    ).with_value(default)


def shape(
    match: Callable[[str], bool],
    *,
    default: Optional[str] = None,
    format: Callable[[str], T] = str,
    help: str = "",
    name: str = "",
    syntax: str = "",
    examples: List[str] = [],
) -> Arg[T, None]:
    return Arg[T, None](
        match=match,
        priority=10,
        duplicates=DuplicatePolicy.SKIP,
        format=format,
        help=help,
        help_template=ShapeArg(
            syntax=syntax,
            examples=examples,
        ),
        name=name,
    ).with_value(default)


def flag(
    match: List[str],
    default: Optional[bool] = False,
    help: str = "",
) -> Arg[bool, None]:
    def check_match(val: str):
        if val in match:
            return True
        return False

    if default:
        raw_value = match[0]
    elif default is None:
        raw_value = None
    else:
        raw_value = ""

    return Arg[bool, None](
        match=check_match,
        priority=20,
        duplicates=DuplicatePolicy.REPLACE,
        format=bool,
        help=help,
        name=", ".join(match),
    ).with_value(raw_value)


def keyword(
    match: List[str],
    *,
    default: Optional[List[str]] = [],
    validate: Callable[[str], T] = str,
    num: int = 1,
    help: str = "",
    lazy: bool = ...,
) -> Arg[bool, T]:
    def check_match(val: str):
        if val in match:
            return True
        return False

    if default:
        raw_value = match[0]
    elif default is None:
        raw_value = None
    else:
        raw_value = ""

    return (
        Arg[bool, T](
            match=check_match,
            priority=20,
            duplicates=DuplicatePolicy.REPLACE,
            format=bool,
            help=help,
            num=num,
            validate=validate,
            greediness=5 if lazy else 30,
            name=", ".join(match) + " <value>",
        )
        .with_value(raw_value)
        .with_values(default or [])
    )



from __future__ import absolute_import

import json
import re
from collections import UserDict
from pathlib import Path

from kslurm.appconfig import Config
from kslurm.args.command import CommandError


class MissingPipdirError(CommandError):
    pass


class KpyIndex(UserDict[str, str]):
    def __init__(self, slurm_tmpdir: Path):
        self._path = slurm_tmpdir / "tmp" / "kpy-index.json"
        if not self._path.exists():
            self._path.parent.mkdir(parents=True, exist_ok=True)
            with self._path.open("w") as f:
                json.dump({}, f)
                self.data = {}
        else:
            with self._path.open("r") as f:
                self.data = json.load(f)

    def write(self):
        with self._path.open("w") as f:
            json.dump(self.data, f)

    def __str__(self):
        return "• " + "\n• ".join(self.data.keys()) if self.data else ""


class VenvCache(UserDict[str, Path]):
    def __init__(self):
        pipdir = Config().get("pipdir")
        if pipdir is None:
            raise MissingPipdirError(
                "pipdir not set. Please set pipdir using `kslurm config pipdir "
                "<directory>`, typically to a project-space or permanent storage "
                "directory"
            )
        self.venv_cache = Path(pipdir, "venv_archives")
        self.venv_cache.mkdir(exist_ok=True)
        venvs_re = [
            re.search(r"(.+)\.tar\.gz$", str(f.name)) for f in self.venv_cache.iterdir()
        ]
        venvs = [r.group(1) for r in venvs_re if r]
        self.data = {v: self._construct_path(v) for v in venvs}

    def get_path(self, name: str):
        try:
            return self.data[name]
        except KeyError:
            return self._construct_path(name)

    def _construct_path(self, name: str):
        return (self.venv_cache / name).with_suffix(".tar.gz")

    def __str__(self):
        return "• " + "\n• ".join(self.data.keys()) if self.data else ""



from __future__ import absolute_import

import itertools as it
from collections import defaultdict
from typing import Any, DefaultDict, Dict, Iterable, List, TypeVar, Union, cast

import attr

from kslurm.args.arg import Arg

T = TypeVar("T")
C = TypeVar("C", bound=Arg[Any, Any])

ModelType = Union[dict[str, Arg[Any, Any]], type]


def group_by_type(items: Iterable[T]) -> Dict[type, List[T]]:
    groupedDict = cast(DefaultDict[type, List[T]], defaultdict(list))
    for key, value in it.groupby(items, type):
        groupedDict[key] += list(value)
    return groupedDict


def get_arg_list(models: ModelType):
    if isinstance(models, dict):
        return [arg.with_id(name) for name, arg in models.items()]
    elif attr.has(models):
        fields = attr.fields(models)
        result: list[Arg[Any, Any]] = []
        for field in fields:
            if field.default is attr.NOTHING or not field.default:
                raise Exception()
            if field.type is attr.NOTHING:
                raise TypeError(f"{field.name} must be annotated")
            if field.default.num and not getattr(field.type, "__origin__") == list:
                raise TypeError(f"{field.name} must be annotated as a list")
            extra_args = {
                **(
                    {"format": field.type}
                    if field.default.format is str and not field.default.num
                    else {}
                ),
                **(
                    {"validate": getattr(field.type, "__args__")[0]}
                    if field.default.validate is str and field.default.num
                    else {}
                ),
            }
            result.append(attr.evolve(field.default, id=field.name, **extra_args))
        return result
    else:
        raise Exception(f"{type(models)} is not a supported object for arg models")


def update_model(
    arg_updates: Iterable[Arg[Any, Any]], model: ModelType
) -> list[Arg[Any, Any]]:
    update = {arg.id: arg for arg in arg_updates}
    if isinstance(model, dict):
        d = {**model, **update}
    elif attr.has(model):
        d = attr.asdict(model(**update), recurse=False)
    else:
        raise Exception(f"{type(model)} is not a supported object for arg models")
    return [arg.with_id(name) for name, arg in d.items()]


def finalize_model(arg_updates: Iterable[Arg[Any, Any]], model: ModelType) -> Any:
    update = {arg.id: (arg.values if arg.num else arg.value) for arg in arg_updates}
    if isinstance(model, dict):
        return cast(T, {**model, **update})
    elif attr.has(model):
        return model(**update)
    else:
        raise Exception(f"{type(model)} is not a supported object for arg models")


def get_help(model: Any) -> bool:
    if attr.has(model):
        help = getattr(model, "help", None)
        return getattr(help, "value", False)
    if isinstance(model, dict):
        help: Any = model.get("help", None)  # type: ignore
        return getattr(help, "value", False)
    raise Exception(f"{type(model)} is not a supported type for arg models")



from __future__ import absolute_import

import importlib.resources as impr
import re
import signal
import subprocess
import subprocess as sp
import sys
from typing import Union

import attr

import kslurm.text as txt
from kslurm.args import keyword
from kslurm.args.command import ParsedArgs, command
from kslurm.exceptions import TemplateError
from kslurm.models.slurm import SlurmModel
from kslurm.slurm.slurm_command import SlurmCommand
from kslurm.style.console import console
from kslurm.venv import VenvCache


@attr.frozen
class JupyterModel(SlurmModel):
    venv: list[str] = keyword(["--venv"])


@command(terminate_on_unknown=True)
def kjupyter(
    args: Union[JupyterModel, TemplateError],
    command_args: list[str],
    arglist: ParsedArgs,
):
    """Start a Jupyter session

    Begins a Jupyter session on an interactive node. By default, it requests 3hr. No
    more time than this should be requested, as this could significantly delay the
    start of the server.

    The --venv arg can be used to specify saved venv. This venv must already
    have jupyter-lab installed, or the command will fail. This will install the
    venv into local scratch, and will thus lead to the best performance.
    Alternatively, you can run the in an already activated virtualenv containing
    jupyter-lab.
    """
    slurm = SlurmCommand(args, command_args, arglist)
    assert isinstance(args, JupyterModel)

    if args.venv:
        venv_cache = VenvCache()
        if args.venv[0] not in venv_cache:
            print("Valid venvs:\n" + str(venv_cache))
            return 1

    with impr.path("kslurm.bin", "kpy-wrapper.sh") as path:
        cmd = [
            "jupyter-lab",
            "--ip",
            "$(hostname -f)",
            "--no-browser",
        ]
        print(txt.KRUN_CMD_MESSAGE.format(args=slurm.slurm_args, command=" ".join(cmd)))
        slurm.command = [
            *(
                [
                    "source",
                    str(path),
                    "&&",
                    "kpy",
                    "load",
                    args.venv[0],
                    "&&",
                ]
                if args.venv
                else []
            ),
            *cmd,
        ]

    if not slurm.test:

        proc = subprocess.Popen(
            slurm.run,
            shell=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        jobid = "0"

        def signal_handler(*args):  # type: ignore
            subprocess.run(f"scancel {jobid}", shell=True)
            sys.exit(0)

        signal.signal(signal.SIGINT, signal_handler)  # type: ignore

        while proc.poll() is None:
            if proc.stdout:
                line = proc.stdout.readline().decode().strip()
                queued = re.match(
                    r"^srun: job (\d+) queued and waiting for resources", line
                )
                url = re.match(
                    r"^https?:\/\/((?:[\w-]+\.)+(?:[\w]+)):(\d+)(\/lab\?token=\w+)$",
                    line,
                )
                if queued:
                    jobid = queued.group(1)
                elif any(
                    [
                        re.match(r"^srun: job \d+ has been allocated resources", line),
                        re.match(
                            r"^To access the server, open this file in a browser:$",
                            line,
                        ),
                        re.match(r"^file:\/\/\/(?:[\w.\-_]+\/)+[\w.\-_]+\.html$", line),
                        re.match(r"^Or copy and paste one of these URLs:$", line),
                        re.match(
                            r"^or http:\/\/(?:[\d]+\.)+\d+:\d+\/lab\?token=\w+$", line
                        ),
                    ]
                ):
                    pass
                elif url:
                    domain = url.group(1)
                    port = url.group(2)
                    path = url.group(3)
                    try:
                        hostname = sp.run(
                            ["wget", "-qO-", "ipinfo.io/ip"], capture_output=True
                        ).stdout.decode().strip()
                    except RuntimeError:
                        hostname = "<hostname>"
                    console.print(
                        txt.JUPYTER_WELCOME.format(
                            port=port,
                            domain=domain,
                            path=path,
                            url=url.group(0),
                            username=sp.run(
                                ["whoami"], capture_output=True
                            ).stdout.decode().strip(),
                            hostname=hostname,
                        )
                    )
                else:
                    print(line)


if __name__ == "__main__":
    kjupyter()



from __future__ import absolute_import, annotations

import abc
from enum import Enum, auto
from typing import (
    Callable,
    DefaultDict,
    Generic,
    Iterable,
    List,
    Optional,
    TypeVar,
    Union,
)

import attr
from colorama import Fore, Style
from rich.table import Table
from rich.text import Text

from kslurm.exceptions import MandatoryArgError, ValidationError

T = TypeVar("T")
S = TypeVar("S")

HelpRow = Union[list[Union[Text, str]], list[Text], list[str]]


class AbstractHelpEntry(abc.ABC):
    title: str
    header: list[str]

    @abc.abstractproperty
    def usage(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def row(self) -> list[Union[Text, str]]:
        raise NotImplementedError()


class AbstractHelpTemplate(abc.ABC):
    title: str
    header: list[str]
    cls_usage: str = ""
    rows: dict[str, list[HelpRow]] = DefaultDict(list)
    priority: int = 0
    right_align_cols: int = 1

    def usage(self, name: str, help: str, default: Optional[str]) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def row(
        self, name: str, help: str, default: Optional[str]
    ) -> Union[list[HelpRow], HelpRow]:
        raise NotImplementedError()

    def add_row(self, name: str, help: str, default: Optional[str]) -> None:
        new_row = self.row(name, help, default)
        if new_row and all([isinstance(row, list) for row in new_row]):
            self.rows[self.__class__.__name__].extend(new_row)  # type: ignore
        elif new_row and all([not isinstance(row, list) for row in new_row]):
            self.rows[self.__class__.__name__].append(new_row)  # type: ignore
        else:
            raise TypeError(
                f"row() in {self} must not return mixture of lists and items"
            )

    @classmethod
    def table(cls):
        t = Table.grid(padding=(0, 2), expand=True)
        for _ in range(cls.right_align_cols):
            t.add_column(justify="right", no_wrap=True)
        if cls.header:
            t.add_row(*cls.header, style="bold")
        for r in cls.rows[cls.__name__]:
            t.add_row(*list(r))
        return t


class DuplicatePolicy(Enum):
    SKIP = auto()
    REPLACE = auto()
    APPEND = auto()


@attr.frozen(eq=False)
class Arg(Generic[T, S]):

    priority: int
    match: Callable[[str], bool]
    duplicates: DuplicatePolicy
    format: Callable[[str], T]
    id: str = ""
    raw_value: str = ""
    help: str = ""
    help_template: Optional[AbstractHelpTemplate] = None
    terminal: bool = False
    name: str = ""
    num: int = 0
    greediness: int = 0
    validate: Callable[[str], S] = lambda x: x
    values: List[S] = []
    values_name: str = ""
    validation_err: Optional[ValidationError] = None
    updated: bool = False

    _value: Optional[T] = None

    @property
    def value(self) -> T:
        if self._value is None:
            raise MandatoryArgError(f"{self.label} has not been provided a value.")
        return self._value

    @property
    def safe_value(self) -> Optional[T]:
        return self._value

    def with_value(self, value: Optional[str], updated: bool = False):
        if value is None:
            return attr.evolve(
                self, value=value, raw_value="", values=[], updated=updated
            )

        try:
            return attr.evolve(
                self,
                value=self.format(value),
                raw_value=value,
                values=[],
                updated=updated,
            )
        except ValidationError as err:
            return attr.evolve(
                self,
                updated=updated,
                validation_err=ValidationError(
                    f"""
    {Fore.RED + Style.BRIGHT}ERROR:{Style.RESET_ALL}
        Invalid value for "{Style.BRIGHT + self.label + Style.RESET_ALL}":
            {err.msg}
            """
                ),
            )

    def with_values(self, values: Iterable[str], updated: bool = False):
        try:
            return attr.evolve(
                self, values=[self.validate(value) for value in values], updated=updated
            )
        except ValidationError as err:
            return attr.evolve(
                self,
                updated=updated,
                validation_err=ValidationError(
                    f"""
    {Fore.RED + Style.BRIGHT}ERROR:{Style.RESET_ALL}
        Invalid value for "{Style.BRIGHT + self.id + Style.RESET_ALL}":
            {err.msg}
            """,
                ),
            )

    def with_err(self, err: ValidationError):
        return attr.evolve(self, validation_err=err)

    def __repr__(self) -> str:
        values_repr = " : '" + "', '".join(map(str, self.values)) if self.values else ""
        return f"<{self.__class__.__name__}: '{self.id}' = {self._value}{values_repr}'>"

    def __eq__(self, o: object) -> bool:
        if isinstance(o, self.__class__):
            return (
                o._value == self._value and o.values == self.values and o.id == self.id
            )
        else:
            return False

    def with_id(self, id: str):
        return attr.evolve(self, id=id)

    @property
    def label(self) -> str:
        if not self.name:
            return self.id
        return self.name



# pyright: reportPrivateUsage=false, reportUnknownArgumentType=false
from __future__ import absolute_import

from typing import Any, NamedTuple

import hypothesis.strategies as st
from hypothesis import assume, given

import kslurm.args.parser as sc
from kslurm.args.arg import Arg, DuplicatePolicy


def get_match_func(match_list: list[str]):
    def match_func(arg: str):
        return arg in match_list

    return match_func


CliParams = NamedTuple(
    "CliParams",
    [
        ("argv", list[str]),
        ("models", list[Arg[Any, Any]]),
        ("results", list[Arg[Any, Any]]),
        ("tail", list[str]),
    ],
)


@st.composite
def cli_params(draw: st.DrawFn):
    alphabet = st.characters(blacklist_categories=("Cs", "Zs"))
    number = draw(st.integers(min_value=0, max_value=5))
    matches: list[str] = []
    for _ in range(number):
        match = draw(st.text(alphabet=alphabet))
        assume(match not in matches)
        assume(match)
        matches.append(match)

    num_args = [
        draw(st.one_of(st.just(0), st.integers(min_value=0, max_value=5)))
        for _ in range(number)
    ]
    priorities = [draw(st.integers(min_value=0, max_value=20)) for _ in range(number)]
    greedinesses = [draw(st.integers(min_value=0, max_value=20)) for _ in range(number)]

    models: list[Arg[Any, Any]] = []
    for i in range(len(matches)):
        model: Arg[Any, Any] = Arg(
            id=f"{matches[i]}/{priorities[i]} (+{num_args[i]}/{greedinesses[i]})",
            priority=priorities[i],
            greediness=greedinesses[i],
            duplicates=DuplicatePolicy.REPLACE,
            match=get_match_func([matches[i]]),
            format=bool,
            num=num_args[i],
        ).with_value("")
        models.append(model)

    used: list[bool] = [False] * number
    argv: list[str] = []
    tail: list[str] = []
    results: list[Arg[Any, Any]] = []
    prev_priority = 0
    for _ in range(draw(st.integers(min_value=0, max_value=number)) * 2):
        higher_priority = [
            match for i, match in enumerate(matches) if priorities[i] >= prev_priority
        ]
        if not len(higher_priority):
            break
        index = matches.index(draw(st.sampled_from(higher_priority)))
        argv.append(matches[index])
        used[index] = True
        num_params = draw(st.integers(min_value=0, max_value=num_args[index]))
        params: list[str] = []
        for _ in range(num_params):
            lower_priority = [
                match
                for i, match in enumerate(matches)
                if priorities[i] < greedinesses[index]
            ]
            higher_priority = [
                match for match in matches if match not in lower_priority
            ]
            if len(lower_priority):
                param = draw(
                    st.one_of(
                        st.text(alphabet=alphabet), st.sampled_from(lower_priority)
                    )
                )
            else:
                param = draw(st.text(alphabet=alphabet))
            assume(param not in higher_priority)
            params.append(param)
        prev_priority = greedinesses[index]
        argv.extend(params)
        results.append(models[index].with_value(matches[index]).with_values(params))

        if num_params == num_args[index]:
            extra_args = draw(st.integers(min_value=0, max_value=2))
            for _ in range(extra_args):
                extra = draw(st.text(alphabet=alphabet))
                assume(extra not in matches)
                argv.append(extra)
                tail.append(extra)

    for i, u in enumerate(used):
        if not u:
            results.append(models[i])

    return CliParams(argv, models, results, tail)


# def quick_arg(id):
#     return Arg(
#         id=id,
#         priority=10,
#         duplicates=DuplicatePolicy.REPLACE,
#         match=get_match_func(['0']),
#         format=bool,
#         num=1,
#     ).with_value("")
@given(cli_params())
def test_parser(params: CliParams):
    argv, models, results, tail = params
    a_result, a_tail = sc._match_args(argv, models, terminate_on_unknown=False)
    assert results == a_result
    assert tail == a_tail



from __future__ import absolute_import

from typing import Any, Callable, Dict, Optional, TypeVar, overload

import kslurm.args.arg_types as arg_types
from kslurm.args.types import WrappedCommand

T = TypeVar("T")

Subcommand = arg_types.Subcommand


@overload
def positional(
    default: Optional[str] = ...,
    *,
    help: str = ...,
    name: str = ...,
) -> Any:
    ...


@overload
def positional(
    default: Optional[str] = ...,
    *,
    format: Callable[[str], T] = ...,
    help: str = ...,
    name: str = ...,
) -> T:
    ...


def positional(*args: Any, **kwargs: Any):  # type: ignore
    return arg_types.positional(*args, **kwargs)


@overload
def choice(
    match: list[str],
    *,
    default: Optional[str] = None,
    help: str = "",
    name: str = "",
) -> Any:
    ...


@overload
def choice(
    match: list[str],
    *,
    default: Optional[str] = ...,
    format: Callable[[str], T] = ...,
    help: str = ...,
    name: str = ...,
) -> T:
    ...


def choice(*args: Any, **kwargs: Any):  # type: ignore
    return arg_types.choice(*args, **kwargs)


def subcommand(
    commands: Dict[str, WrappedCommand],
    default: Optional[str] = None,
) -> arg_types.Subcommand:
    return arg_types.subcommand(commands, default)  # type: ignore


@overload
def shape(
    match: Callable[[str], bool],
    *,
    default: Optional[str] = None,
    help: str = "",
    name: str = "",
    syntax: str = "",
    examples: list[str] = [],
) -> Any:
    ...


@overload
def shape(
    match: Callable[[str], bool],
    *,
    default: Optional[str] = None,
    format: Callable[[str], T] = str,
    help: str = "",
    name: str = "",
    syntax: str = "",
    examples: list[str] = [],
) -> T:
    ...


def shape(  # type: ignore
    *args: Any,
    **kwargs: Any,
):
    return arg_types.shape(*args, **kwargs)


def flag(
    match: list[str],
    default: Optional[bool] = False,
    help: str = "",
) -> bool:
    return arg_types.flag(match, default, help)  # type: ignore


@overload
def keyword(
    match: list[str],
    *,
    default: list[str] = ...,
    num: int = ...,
    lazy: bool = ...,
    help: str = ...,
) -> Any:
    ...


@overload
def keyword(
    match: list[str],
    *,
    default: list[str] = ...,
    num: int = ...,
    validate: Callable[[str], T] = ...,
    lazy: bool = ...,
    help: str = ...,
) -> list[T]:
    ...


def keyword(  # type: ignore
    *args: Any,
    **kwargs: Any,
):
    return arg_types.keyword(*args, **kwargs)



from __future__ import absolute_import

import itertools as it
import operator as op
from pathlib import Path
from typing import Any, DefaultDict, Optional, Union

from rich.console import Group
from rich.padding import Padding
from rich.table import Table
from rich.text import Text

from kslurm.args.arg import AbstractHelpTemplate, Arg
from kslurm.style.console import console


class BasicTemplate(AbstractHelpTemplate):
    title = "options"
    header = []
    cls_usage = "<options>"
    priority = 50

    def row(self, name: str, help: str, default: Optional[str]):
        return [name, help, default if default is not None else ""]


def _header(header: str):
    return f"[bold green]{header.upper()}[/bold green]"


def _section(header: str, body: Union[Table, str]):
    return (
        Group(Padding.indent(_header(header), 4), Padding.indent(body, 8), "\n")
        if body
        else ""
    )


def print_help(
    script: str,
    models: list[Arg[Any, Any]],
    script_help: str = "",
    usage_suffix: str = "",
) -> None:
    script_name = Path(script).name

    templates: set[type[AbstractHelpTemplate]] = set()
    usages: dict[type[AbstractHelpTemplate], list[str]] = DefaultDict(list)
    for model in models:
        if model.help_template is None:
            template = BasicTemplate()
        else:
            template = model.help_template
        args = model.label, model.help, str(model.safe_value or "")
        template.add_row(*args)
        try:
            usages[type(template)].append(template.usage(*args))
        except NotImplementedError:
            pass
        templates.add(type(template))

    sorted_templates: dict[type[AbstractHelpTemplate], list[str]] = {}

    for template in sorted(usages, key=op.attrgetter("priority"), reverse=True):
        if not usages[template]:
            usages[template].append(template.cls_usage)
        sorted_templates[template] = usages[template]

    command_line_example = [
        "[b]USAGE:[/] ",
        script_name,
        " " + " ".join(it.chain.from_iterable(sorted_templates.values()))
        if usages
        else "",
        Text(f" <{usage_suffix}>", style="cyan") if usage_suffix else "",
        "\n",
    ]

    sections = [_section(template.title, template.table()) for template in templates]

    console.print(*command_line_example, script_help, *sections, sep="")



# from __future__ import absolute_import

# import re

from __future__ import absolute_import

# from kslurm.args.arg_types import flag, keyword, positional, shape
# from kslurm.args.arg import Arg

# def time_format(time: str):
#     if "-" in time:
#         days, hhmm = time.split("-")
#         hours, min = hhmm.split(":")
#     else:
#         days = 0
#         hours, min = time.split(":")
#     return str(int(min) + int(hours) * 60 + int(days) * 60 * 24)


# class Templates:
#     time: Arg[str, None] = shape(
#         match=lambda arg: bool(re.match(r"^([0-9]{1,2}-)?[0-9]{1,2}:[0-9]{2}$", arg)),
#         format=time_format,
#         default="03:00",
#     )

#     gpu: Arg[bool, None] = flag(match=["gpu"])

#     number: Arg[int, None] = shape(
#         match=lambda arg: bool(re.match(r"^[0-9]+$", arg)), default="1"
#     )

#     job_template: Arg[bool, str] = keyword(match=["-j", "--job-template"], num=1)

#     length_5_keyword: Arg[bool, str] = keyword(match=["--length_5_keyword"], num=5)

#     lazy_inf_keyword: Arg[bool, str] = keyword(
#         match=["lazy_inf_keyword"],
#         num=-1,
#         lazy=True,
#     )

#     greedy_inf_keyword: Arg[bool, str] = keyword(
#         match=["greedy_inf_keyword"],
#         num=-1,
#     )

#     positional1: Arg[str, None] = positional("pos1")
#     positional2: Arg[str, None] = positional("pos2")
#     positional3: Arg[str, None] = positional("pos3")
#     positional4: Arg[str, None] = positional("pos4")
#     positional5: Arg[str, None] = positional("pos5")



__submodules__ = ["kbatch", "kpy", "krun", "kjupyter"]

# <AUTOGEN_INIT>
from kslurm.cli.kbatch import (
    kbatch,
)
from kslurm.cli.kpy import (
    MissingPipdirError,
    MissingSlurmTmpdirError,
    VenvCache,
    kpy,
)
from kslurm.cli.krun import (
    krun,
)
from kslurm.cli.kjupyter import (
    kjupyter,
)

__all__ = [
    "MissingPipdirError",
    "MissingSlurmTmpdirError",
    "VenvCache",
    "kbatch",
    "kjupyter",
    "kpy",
    "krun",
]

# </AUTOGEN_INIT>



from __future__ import absolute_import

from kslurm.args.arg import Arg
from kslurm.args.arg_types import positional


def positional_models(num: int) -> list[Arg[str, None]]:
    return [positional(f"pos{num}") for _ in range(num)]



from __future__ import absolute_import

import functools as ft
import inspect
import sys
from typing import Any, Callable, Literal, Optional, Union, overload

import attr

from kslurm.args.arg import Arg
from kslurm.args.arg_types import flag
from kslurm.args.helpers import finalize_model, get_arg_list
from kslurm.args.parser import ScriptHelp, parse_args
from kslurm.args.types import WrappedCommand
from kslurm.exceptions import CommandLineError

_CommandFunc = Callable[..., Optional[int]]
ModelType = Union[dict[str, Arg[Any, Any]], type]


ParsedArgs = dict[str, Arg[Any, Any]]


class CommandError(Exception):
    def __init__(self, msg: str, *args: Any):
        super().__init__(*args)
        self.msg = msg


class CommandArgs:
    _model: Optional[str] = None
    _tail: Optional[str] = None
    _modellist: Optional[str] = None
    _name: Optional[str] = None

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value: str):
        if self._model is not None:
            raise ValueError("model is already set")
        self._model = value

    @property
    def tail(self):
        return self._tail

    @tail.setter
    def tail(self, value: str):
        if self._tail is not None:
            raise ValueError("tail is already set")
        self._tail = value

    @property
    def modellist(self):
        return self._modellist

    @modellist.setter
    def modellist(self, value: str):
        if self._modellist is not None:
            raise ValueError("modellist is already set")
        self._modellist = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value: str):
        if self._name is not None:
            raise ValueError("name is already set")
        self._name = value


@overload
def command(
    maybe_func: None = ...,
    *,
    terminate_on_unknown: bool = ...,
    typer: bool = ...,
) -> Callable[[_CommandFunc], WrappedCommand]:
    ...


@overload
def command(
    maybe_func: _CommandFunc = ...,
    *,
    terminate_on_unknown: Literal[True] = ...,
    typer: Literal[False] = ...,
) -> WrappedCommand:
    ...


def command(
    maybe_func: Optional[_CommandFunc] = None,
    *,
    terminate_on_unknown: bool = False,
    typer: bool = False,
) -> Union[WrappedCommand, Callable[[_CommandFunc], WrappedCommand]]:
    @attr.frozen
    class BlankModel:
        pass

    def decorator(func: _CommandFunc):
        if not callable(func):
            raise CommandLineError(f"{func} is not callable")

        params = inspect.signature(func)
        exceptions: Any = tuple()
        model = BlankModel

        command_args = CommandArgs()
        if typer:
            model = attr.make_class(
                "__Typer_Model__",
                {
                    param.name: attr.attrib(
                        default=param.default
                        if param.default is not params.empty
                        else attr.NOTHING,
                        type=param.annotation
                        if param.annotation is not params.empty
                        else attr.NOTHING,
                    )
                    for param in params.parameters.values()
                },
            )
        else:
            for param in params.parameters.values():
                if param.annotation == param.empty:
                    raise CommandLineError(
                        f"parameter '{param}' in {func} must be annotated"
                    )
                if param.annotation == list[str]:
                    command_args.tail = param.name
                    continue
                if param.annotation == ParsedArgs:
                    command_args.modellist = param.name
                    continue
                if param.annotation == str:
                    command_args.name = param.name
                    continue

                if model is not BlankModel:
                    raise CommandLineError(
                        f"Mutliple models detected: {model} and {param.annotation}"
                    )
                model = param.annotation
                if isinstance(model, str):
                    model = eval(model, func.__globals__)

                types = getattr(model, "__args__", None)
                if types is not None:
                    model = types[0]
                    exceptions = types[1:]

                if not attr.has(model):
                    raise CommandLineError(
                        f"Annotation of {param} in {func} must be an attr "
                        f"annotated class (currently {model})"
                    )
                command_args.model = param.name

        @ft.wraps(func)
        def wrapper(argv: list[str] = sys.argv):
            doc = func.__doc__
            if doc is None:
                doc = ""

            model_list = get_arg_list(model)
            model_list.append(flag(match=["--help", "-h"]).with_id("help"))
            try:
                if command_args.tail:
                    h = ScriptHelp(
                        script_name=argv[0], docstring=doc, usage_suffix="command_args"
                    )
                else:
                    h = ScriptHelp(script_name=argv[0], docstring=doc)
                parsed_list, tail = parse_args(
                    argv[1:],
                    model_list,
                    help=h,
                    terminate_on_unknown=terminate_on_unknown,
                )
                parsed_list = list(filter(lambda arg: arg.id != "help", parsed_list))
                parsed = finalize_model(parsed_list, model)
                if typer:
                    args = attr.asdict(parsed, recurse=False)
                else:
                    args = {
                        **(
                            {command_args.model: parsed}
                            if command_args.model is not None
                            else {}
                        ),
                        **(
                            {command_args.tail: tail}
                            if command_args.tail is not None
                            else {}
                        ),
                        **(
                            {
                                command_args.modellist: {
                                    arg.id: arg for arg in parsed_list
                                }
                            }
                            if command_args.modellist is not None
                            else {}
                        ),
                        **(
                            {command_args.name: argv[0]}
                            if command_args.name is not None
                            else {}
                        ),
                    }
            except exceptions as err:
                parsed = err
                args = {
                    **(
                        {command_args.model: parsed}
                        if command_args.model is not None
                        else {}
                    ),
                }
            try:
                return func(**args) or 0
            except CommandError as err:
                print(err.msg)
                return 1

        return wrapper

    if maybe_func is None:
        return decorator
    return decorator(maybe_func)



# from __future__ import absolute_import

# from typing import Any, List

from __future__ import absolute_import

# from kslurm.args.arg import Arg
# from .arg_templates import Templates


# class DummyArgs:

#     ###
#     # Args used to test keyword groupping
#     two_keyword: List[Arg[Any, Any]] = [
#         Templates.time,
#         Templates.gpu,
#         Templates.job_template,
#         Templates.positional1,
#         Templates.time,
#         Templates.length_5_keyword,
#         Templates.positional1,
#         Templates.time,
#         Templates.gpu,
#         Templates.positional1,
#         Templates.positional1,
#         Templates.gpu,
#         Templates.number,
#         Templates.time,
#     ]

#     positional_patchwork: List[Arg[Any, Any]] = [
#         Templates.positional1,
#         Templates.positional2,
#         Templates.positional3,
#     ]

#     positional_wrapped_keyword: List[Arg[Any, Any]] = [
#         Templates.positional1,
#         Templates.job_template,
#         Templates.positional1,
#         Templates.positional1,
#     ]

#     lazy_keywords: List[Arg[Any, Any]] = [
#         Templates.positional1,
#         Templates.lazy_inf_keyword,
#         Templates.positional1,
#         Templates.positional1,
#         Templates.lazy_inf_keyword,
#         Templates.positional1,
#         Templates.positional1,
#         Templates.positional1,
#         Templates.positional1,
#         Templates.time,
#     ]

#     greedy_keywords: List[Arg[Any, Any]] = [
#         Templates.greedy_inf_keyword,
#         Templates.lazy_inf_keyword,
#         Templates.positional1,
#         Templates.time,
#     ]


# class ArgStr:
#     positionals: List[str] = ["Hello", "beautiful", "world"]

#     extra_positionals: List[str] = [
#         "Hello",
#         "beautiful",
#         "world",
#         "Here",
#         "be",
#         "extra",
#     ]

#     two_keyword: List[str] = [
#         "07:23",
#         "gpu",
#         "--job-template",
#         "parcour",
#         "3",
#         "--length_5_keyword",
#         "one",
#         "of",
#         "the",
#         "five",
#         "values",
#         "command",
#     ]

#     shapes_and_positionals: List[str] = [
#         "07:23",
#         "gpu",
#         "--job-template",
#         "parcour",
#         "one",
#         "of" "3",
#         "the",
#         "five",
#         "values",
#         "command",
#     ]



from __future__ import absolute_import

import subprocess
from typing import Union

from colorama import Fore

import kslurm.text as txt
from kslurm.args.command import ParsedArgs, command
from kslurm.exceptions import TemplateError
from kslurm.models.slurm import SlurmModel
from kslurm.slurm.slurm_command import SlurmCommand


@command(terminate_on_unknown=True)
def kbatch(
    args: Union[SlurmModel, TemplateError],
    command_args: list[str],
    arglist: ParsedArgs,
):
    """Submit a job using sbatch

    Supports scripts (e.g. ./script.sh) or direct commands (e.g. cp dir/file.txt dir2)

    When your command contains bash interpreted elements such as $VARIABLES and
    $(subshells), these will be immediately expanded. Normally, this behaviour
    is fine, but sometimes they should only be interpretted on the allocated cluster.
    For instance, $SLURM_TMPDIR will evaluate to "" unless it is interpretted later.
    To force this behaviour, wrap the $VARIABLE or $(subshell) in quotes:
        '$SLURM_TMPDIR'
        '$(hostname)'
    """

    slurm = SlurmCommand(args, command_args, arglist)
    command = slurm.command if slurm.command else f"{Fore.RED}Must provide a command"

    print(txt.KBATCH_MSG.format(slurm_args=slurm.slurm_args, command=command))
    if slurm.command:
        proc = subprocess.run(
            slurm.batch, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )

        out = proc.stdout.decode()

        if proc.returncode != 0:
            print(Fore.WHITE + out)
        if slurm.test:
            # output will be the issued command, so we print it
            print(Fore.WHITE + out)
        else:
            # We subtract the last character of the output
            # to remove the final "\n" character and get the
            # job_id
            slurmid = out[:-1]

            print(
                f"""Scheduled job {slurmid}.
    To cancel, run:
        scancel {slurmid}
        """
            )


if __name__ == "__main__":
    kbatch(["kbatch", "--help"])



__submodules__ = ["arg_types_cast"]

__ignore__ = ["T", "S"]

# <AUTOGEN_INIT>
from kslurm.args.arg_types_cast import (
    Subcommand,
    choice,
    flag,
    keyword,
    positional,
    shape,
    subcommand,
)

__all__ = [
    "Subcommand",
    "choice",
    "flag",
    "keyword",
    "positional",
    "shape",
    "subcommand",
]

# </AUTOGEN_INIT>



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kslurm',
 'kslurm.args',
 'kslurm.bin',
 'kslurm.cli',
 'kslurm.data',
 'kslurm.installer',
 'kslurm.models',
 'kslurm.slurm',
 'kslurm.style',
 'kslurm.test',
 'kslurm.test.args',
 'kslurm.test.args.models',
 'kslurm.test.installer',
 'neuroglia_helpers',
 'neuroglia_helpers.etc']

package_data = \
{'': ['*'],
 'neuroglia_helpers': ['.github/workflows/*',
                       '.github_scripts/*',
                       'bin/*',
                       'cfg/*',
                       'cfg_apps/*',
                       'job-templates/*']}

install_requires = \
['InquirerPy>=0.3.3,<0.4.0',
 'appdirs>=1.4.4,<2.0.0',
 'attrs>=21.2.0,<22.0.0',
 'colorama>=0.4.4,<0.5.0',
 'docstring-parser>=0.14.1,<0.15.0',
 'rich>=12.2.0,<13.0.0',
 'semver>=2.13.0,<3.0.0',
 'shellingham>=1.4.0,<2.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'typing-extensions>=3.10',
 'virtualenv>=20.0.24,<21']

entry_points = \
{'console_scripts': ['kbatch = kslurm.cli.kbatch:kbatch',
                     'kjupyter = kslurm.cli.kjupyter:kjupyter',
                     'kpy = kslurm.cli.kpy:kpy',
                     'krun = kslurm.cli.krun:krun',
                     'kslurm = kslurm.cli.main:main']}

setup_kwargs = {
    'name': 'kslurm',
    'version': '0.4.0',
    'description': 'Helper scripts and wrappers for running commands on SLURM compute clusters.',
    'long_description': "Utility functions to make working with SLURM easier.\n\n# Installation\nThe recommend way to install kslurm is via [`pipx`](https://pypa.github.io/pipx), a tool for installing python applications.\nThis will make kslurm globally available without infecting your global python environment.\nInstallation instructions for pipx can be found on their [website](https://pypa.github.io/pipx).\nOnce installed, simply run\n```bash\npipx install kslurm\n```\n\nNote that kslurm requires Python 3.9 or higher.\nIf pipx was installed using a lower version (e.g. 3.8), you will need to manually specify the python executable to use.\nActivate the appropriate python version (e.g. `module load python/3.10`) so that when you run `python --version`, the correct version appears.\nThen run\n\n```bash\npipx install kslurm --python $(which python)\n```\n\nFor full kslurm features, including integration with `pip`, you need to source the init script, preferably in your `~.bash_profile` (the init script contains commands that may not be available on non-login nodes). You can do this by running:\n\n```bash\nkpy bash >> $HOME/.bash_profile\n```\n\nFinally, you need to complete some basic configuration. First, set your SLURM account. Run\n\n```bash\nkslurm config account -i\n```\n\nThis will begin an interactive session letting you choose from the accounts available to you. Each account will be listed with it's LevelFS. The higher the LevelFS, the more underused the account is, so prefer accounts with higher values.\n\nNext, set your pipdir. This will be used to store python wheels and virtual envs. It should be in a permanent storage or project directory. For instance, on ComputeCanada servers, it should go in `$HOME/projects/<account_name>/<user_name>/.kslurm`. Use the following command:\n\n```bash\nkslurm config pipdir <dir>\n```\n\n## Upgrading and uninstalling\n\nThe app can be updated by running\n```bash\npipx upgrade kslurm\n```\n\nand removed using\n```bash\npipx uninstall kslurm\n```\n\n## Neuroglia-helpers Integration\n\nSee the [dedicated page](docs/neuroglia-helpers.md).\n\n## Legacy Installer\nkslurm includes an installation script that, previously, was the recommended install method.\nWhile it should technically still work, it is no longer supported and may be removed in the future.\nIts instructions are included, for reference, below.\n\nUsers who previously installed kslurm via this script should switch to a pipx install for long term support. Simply uninstall `kslurm` using the instructions below, then install via pipx as described above\n\nInstallation is via the following command:\n```\ncurl -sSL https://raw.githubusercontent.com/pvandyken/kslurm/master/install_kslurm.py | python -\n```\n\nIf you wish to uninstall, run the same command with `--uninstall` added to the end.\n\nThe package can be updated by running `kslurm update`.\n\n# Features\nCurrently offers four commands:\n* kbatch: for batch submission jobs (no immediate output)\n* krun: for interactive submission\n* kjupyter: for Jupyter sessions\n* kpy: for python environment management\n\nAll three use a regex-based argument parsing, meaning that instead of writing a SLURM file or supplying confusing `--arguments`, you can request resources with an intuitive syntax:\n\n```\nkrun 4 3:00 15G gpu\n```\nThis command will request an interactive session with __4__ cores, for __3hr__, using __15GB__ of memory, and a __gpu__.\n\nAnything not specfically requested will fall back to a default. For instance, by default the commands will request 3hr jobs using 1 core with 4GB of memory. You can also run a predefined job template using -j _template_. Run either command with -J to get a list of all templates. Any template values can be overriden simply by providing the appropriate argument.\n\nThe full list of possible requests, their syntaxes, and their defaults can be found at the bottom of the README.\n\n## krun\n\nkrun is used for interactive sessions on the cluster. If you run krun all by itself, it will fire up an interactive session on the cluster:\n\n```\nkrun\n```\nYou'll notice the server name in your terminal prompt will be changed to the cluster assigned to you. To end the session, simply use the `exit` command.\n\nYou can also submit a specific program to run:\n\n```\nkrun 1:00 1G python my_program.py\n```\nThis will request a 1hr session with one core and 1 GB of memory. The output of the job will be displayed on the console. Note that your terminal will be tied to the job, so if you quit, or get disconnected, your job will end. (tmux can be used to help mitigate this, see this [tutorial from Yale](https://docs.ycrc.yale.edu/clusters-at-yale/guides/tmux/) for an excellent overview).\n\nNote that you should never request more than the recommended amount of time for interactive jobs as specified by your cluster administrator. For ComputeCanada servers, you should never request more than 3 hr. If you do, you'll be placed in the general pool for resource assignment, and the job could take hours to start. Jobs of 3hr or less typically start in less than a minute.\n\n## kbatch\n\nJobs that don't require monitoring of the output or immediate submission, or will run for more than three hours, should be submitted using `kbatch`. This command schedules the job, then returns control of the terminal. Output from the job will be placed in a file in your current working directory entitled `slurm-[jobid].out`.\n\nImproving on `sbatch`, `kbatch` does not require a script file. You can directly submit a command:\n\n```\nkbatch 2-00:00 snakemake --profile slurm\n```\nThis will schedule a 2 day job running snakemake.\n\nOf course, complicated jobs can still be submitted using a script. Note that kbatch explictely specifies the resources it knows about in the command line. Command line args override `#SBATCH --directives` in the submit script, so at this time, you cannot use such directives to request resources unless they are not currently supported by kslurm. This may change in a future release.\n\n## kjupyter\n\nThis command requests an interactive job running a jupyter server. As with krun, you should not request a job more than the recommended maximum time for your cluster (3hr for ComputeCanada). If you need more time than that, just request a new job when the old one expires.\n\nYou should not provide any extra command to kjupyter. Just supply whatever resources you wish to request.\n```\nkjupyter 32G 2\n```\nThis will start a jupyter session with 32 GB of memory and 2 cores.\n\nNote that the command will fail if there is no `jupyter-lab` executable on the `PATH`. Use `pip install jupyterlab` if it's not installed. Typically, you should do this within a Python environment using a tool of your choice (e.g. virtualenv).\n\n# Unsupported SLURM args\n\nCurrently, the only way to supply arguments to SLURM beyond the items listed below is to list it as an `#SBATCH --directive` in a submission script. This only works with `kbatch`, not `krun` or `kjupyter`. A future release may support a method to supply these arguments directly on the command line. If you frequently use an option not listed below, make an issue and we can discuss adding support!\n\n# Slurm Syntax\n\nThe full syntax is outlined below. You can always run a command with `-h` to get help.\n\n| Resource  |           Syntax           |                                                                   Default |                                                    Description |\n| :-------- | :------------------------: | ------------------------------------------------------------------------: | -------------------------------------------------------------: |\n| Time      | [d-]dd:dd -> [days-]hh:mm  |                                                                       3hr |                                   The amount of time requested |\n| CPUS      |     d -> just a number     |                                                                         1 |                                   The number of CPUs requested |\n| Memory    |            d(G/M)[B] -> e.g. 4G, 500MB |                                                            4GB | The amount of memory requested |\n| Account   | --account <_account name_> |                                                                           |                      The account under which to submit the job. A default account can be configured using `kslurm config account <account_name>` |\n| GPU       |            gpu             |                                                                     False |                         Provide flag to request 1 GPU instance |\n| Directory |  <_any valid directory_>   |                                                                        ./ | Change the current working directory before submitting the job |\n| x11       |           --x11            |                                                                     False |                   Requests x11 forwarding for GUI applications |\n\n# kpy\n\nkpy bundles a set of commands to help manage pip virtual environments on Slurm compute clusters, specifically addressing a few issues unique to such servers:\n\n## Ephemeral venvs\nIn most use cases, python venvs are installed on compute clusters, ideally on local scratch storage.\nThis makes venvs inherently ephemeral.\nBecause installing a venv can take an appreciable amount of time, kpy packs tools to archive entire venvs for storage in a permanent local repository (ideally located in project-specific or permanent storage). Once saved, venvs can be quickly reloaded into a new compute environment.\n\nNote that copying venvs from one location to another is not a trivial task. The current setup has been tested on ComputeCanada servers without any issues so far, but problems may arise on another environment.\n\n## No internet\nCompute clusters often don't have an internet connection, limiting our install repertoire to locally available wheels.\nWith kpy, venvs can be created on a login node (using the available internet connection), then saved and loaded onto a compute node.\nKpy also includes some optional bash tools (see `kpy bash`), including a wrapper around pip that prevents it from accessing the internet on compute nodes, and connecting it with a local private wheelhouse.\n\n## Commands\n### `create`\n\n```bash\n# usage\nkpy create [<version|3.x>] [<name>]\n```\n\nCreate a new environment.\nName is optional; if not provided, a placeholder name will be created.\nVersion must be of the form `3.x` where x is any number (e.g `3.8`, `3.10`).\nIf provided, the corresponding python version will be used in the virtual env.\nNote that an appropriate python executable must be somewhere on your path (e.g. for `3.8` -> `python3.8`).\nIf not provided, the python version used to install kslurm will be used.\n\nIf run on a login node, the env will be created in a `$TMPDIR`.\nIf run on a compute node, it will be created in `$SLURM_TMPDIR`.\n\n### `save`\n\n```bash\n# usage\nkpy save [-f] <name>\n```\n\nSave the venv to your permanent cache.\nThis requires setting `pipdir` in the kslurm config (see below).\nBy default, `save` will not oversave an existing cache, but `-f` can be included to override this behaviour.\nIf a new name is provided, it will be used to update the current venv name and prompt.\n\n### `load`\n\n\n```bash\n# usage\nkpy load [<name>] [--as <newname>]\n```\n\nLoad a saved venv from the cache.\nThis command only works on a compute node (i.e. `$SLURM_TMPDIR` must be defined).\nIf a venv called `<name>` already exists, the command will fail, as each name can only be used once.\n`--as <newname>` works around this by changing the name of the loaded venv (the name of the saved venv will remain the same)\nCalling `load` without any `<name>` will print a list of current cached venvs.\n\n### `activate`\n\n```bash\n# usage\nkpy activate [<name>]\n```\n\nActivate venv initialized using `create` or `load`.\nName will be the same as the name appearing in the venv prompt (i.e. the name provided on initial loading or creation, through `--as`, or the last saved name).\nThis command only works on a compute node.\nVenvs created on a login node cannot be directly activated using kpy.\n\n### `list`\n\n```bash\n# usage\nkpy list\n```\n\nList currently initialized venvs (i.e. venvs you can `activate`).\nThis command only works on a compute node.\n\n\n### `bash`\n\n```bash\n# usage\nkpy bash\n```\n\nEchos a line of bash script that can be added to your `.bashrc` file:\n\n```bash\nkpy bash >> $HOME/.bashrc\n```\n\nThis adds a few features to your command line environment:\n\n- **pip wrapper**: Adds a wrapper around pip that detects if you are on a login node when running `install`, `wheel`, or `download`. If not on a login node, the `--no-index` flag will be appended to the command, preventing the use of an internet connection.\n- **wheelhouse management**: If `pipdir` is configured in the kslurm config, a wheelhouse will be created in your pip repository. Any wheels downloaded using `pip wheel` will be placed in that wheelhouse, and all wheels in the wheelhouse will be discoverable by `pip install`, both on login and compute nodes.\n\n## Configuration\n\nkslurm currently supports a few basic configuration values, and more will come with time. All configuration can be set using the command\n\n```bash\nkslurm config <key> <value>\n```\n\nYou can print the value of a configuration using\n\n```bash\nkslurm config <key>\n```\n\n### Current values\n\n* `account`: Default account to use for kslurm commands (e.g. `kbatch`, `krun`, etc)\n* `pipdir`: Directory to store cached venvs and wheels. Should be a project or permanent storage dir.\n",
    'author': 'Peter Van Dyken',
    'author_email': 'pvandyk2@uwo.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pvandyken/kslurm',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import absolute_import

import re
from typing import Any, List

import attr
from typing_extensions import TypedDict

import kslurm.test.args.models.formatters as formatters
from kslurm.args.arg_types import flag, keyword, shape
from kslurm.test.args.models.common import ModelTest, update_model


def get_tests(model: object) -> List[List[List[Any]]]:
    return [
        [
            [
                "07:23",
                "gpu",
                "--job-template",
                "parcour",
                "3",
                "--length_5_keyword",
                "one",
                "of",
                "the",
                "five",
                "values",
                "command",
            ],
            # Fall back on defaults
            [
                "3",
                "command",
                "of",
                "--length_5_keyword",
                "five",
                "07:23",
                "--job-template",
                "parcour",
                "the",
                "one",
                "values",
                "gpu",
            ],
            # Keyword masking shapes and flags
            [
                "07:23",
                "--length_5_keyword",
                "command",
                "gpu",
                "3",
                "--job-template",
                "parcour",
                "10",
                "of",
                "five",
                "the",
                "one",
                "values",
            ],
        ],
        [
            update_model(
                model,
                [
                    "07:23",
                    True,
                    3,
                    [True, ["parcour"]],
                    [True, ["one", "of", "the", "five", "values"]],
                    [None, ["command"]],
                ],
            ),
            update_model(
                model,
                [
                    None,
                    None,
                    3,
                    None,
                    None,
                    [
                        None,
                        [
                            "command",
                            "of",
                            "--length_5_keyword",
                            "five",
                            "07:23",
                            "--job-template",
                            "parcour",
                            "the",
                            "one",
                            "values",
                            "gpu",
                        ],
                    ],
                ],
            ),
            update_model(
                model,
                [
                    "07:23",
                    None,
                    "10",
                    None,
                    [True, ["command", "gpu", "3", "--job-template", "parcour"]],
                    [None, ["of", "five", "the", "one", "values"]],
                ],
            ),
        ],
    ]


@attr.define
class AttrModel:
    time: str = shape(
        match=lambda arg: bool(re.match(r"^([0-9]{1,2}-)?[0-9]{1,2}:[0-9]{2}$", arg)),
        format=formatters.time,
        default="03:00",
    )

    gpu: bool = flag(match=["gpu"])

    number: int = shape(
        match=lambda arg: bool(re.match(r"^[0-9]+$", arg)), format=int, default="1"
    )

    job_template: list[int] = keyword(match=["-j", "--job-template"], num=1)

    length_5_keyword: list[str] = keyword(
        match=["--length_5_keyword"],
        num=5,
    )


# f = parse_args(["1:30", "gpu", "--job-template", "4", "5"], AttrModel)


class TypedDictModel(TypedDict):
    time: str
    gpu: bool
    number: str
    job_template: str
    length_5_keyword: str


attrmodel = AttrModel()

typed_dict_model = TypedDictModel(
    time=shape(
        match=lambda arg: bool(re.match(r"^([0-9]{1,2}-)?[0-9]{1,2}:[0-9]{2}$", arg)),
        format=formatters.time,
        default="03:00",
    ),
    gpu=flag(match=["gpu"]),
    number=shape(match=lambda arg: bool(re.match(r"^[0-9]+$", arg)), default="1"),
    job_template=keyword(match=["-j", "--job-template"], num=1),
    length_5_keyword=keyword(
        match=["--length_5_keyword"],
        num=5,
    ),
)

basic_attr = ModelTest(attrmodel, *get_tests(attrmodel))

basic_dict = ModelTest(typed_dict_model, *get_tests(typed_dict_model))



from __future__ import absolute_import

import json
from collections import UserDict
from pathlib import Path

import appdirs

CONFIG_PATH = Path(appdirs.user_config_dir("kslurm"), "config.json")


class Config(UserDict[str, str]):
    def __init__(self):
        self._path = CONFIG_PATH
        if not self._path.exists():
            self._path.parent.mkdir(parents=True, exist_ok=True)
            with self._path.open("w") as f:
                json.dump({}, f)
                self.data = {}
        else:
            with self._path.open("r") as f:
                self.data = json.load(f)

    def write(self):
        with self._path.open("w") as f:
            json.dump(self.data, f)

    def get_children(self, entity: str):
        for key, value in self.data.items():
            if key.startswith(entity + "."):
                yield key, value

    def __str__(self):
        return "• " + "\n• ".join(self.data.keys()) if self.data else ""
